/*
 * VFD Clock
 * Kerry Wong
 * 6/2013
 */

#include <Arduino.h>
#include <Wire.h>

#define DS3232_I2C_ADDRESS B01101000  // This is the I2C address 7bits

enum DisplayMode {
    TIME_12,
    TIME_24,
    DATE_DOW
};

enum DateSetupMode {
    NORMAL_DATE_DISPLAY,
    BLINK_MONTH,
    BLINK_DAY,
    BLINK_YEAR,
    BLINK_DOW
};

enum TimeSetupMode {
    NORMAL_TIME_DISPLAY,
    BLINK_HOUR,
    BLINK_MINUTE,
    BLINK_SECOND 
};

typedef struct {
  byte second;
  byte minute;
  byte hour;
  byte dayOfWeek;
  byte dayOfMonth;
  byte month;
  byte year;
} Date;

Date pD;

DisplayMode currentDisplayMode = TIME_12;
DateSetupMode currentDateSetupMode = BLINK_DAY;
TimeSetupMode currentTimeSetupMode = BLINK_MINUTE;

int currentSetupMode = 0;

const int NUM_OF_GRIDS = 13;
const int PIN_B0 = 2;
const int PIN_B1 = 3;
const int PIN_B2 = 4;
const int PIN_B3 = 5;

const int PIN_DS = 6;
const int PIN_OE = 7;
const int PIN_LATCH = 8;
const int PIN_CLK = 9;

//'0'..'9'
const int DIGITS[]={63, 6, 91, 79, 102, 109, 125, 7, 127, 111, 63};
const int DASH = 64;
const int DOT = 128;
const int COMMA = 256;
 
/*
 *  1 2 3 4 5 6 7 8 9 10 11 12 
 *      h h - m m - s s
 *  m m - d d - y y y y    dow
 */

const int HOUR1 = 2;
const int HOUR2 = 3;
const int DASH1 = 4;
const int MINUTE1 = 5;
const int MINUTE2 = 6;
const int DASH2 = 7;
const int SECOND1 = 8;
const int SECOND2 = 9;

const int MONTH1 = 0;
const int MONTH2 = 1;
const int DASH3 = 2;
const int DAY1 = 3;
const int DAY2 = 4;
const int DASH4 = 5;
const int YEAR1 = 6;
const int YEAR2 = 7;
const int YEAR3 = 8;
const int YEAR4 = 9;
const int DOW = 11;

const int BTN_SETUP = 14;
const int BTN_UP = 10;
const int BTN_DOWN = 11;
const int BTN_DISP_MODE = 12;

const int IDX_BTN_DISP_MODE = 0;
const int IDX_BTN_UP = 1;
const int IDX_BTN_DOWN = 2;
const int IDX_BTN_SETUP = 3;

const int BLINK_DELAY = 100;

int buttons[]={BTN_DISP_MODE, BTN_UP, BTN_DOWN, BTN_SETUP};

int buttonReadings[4];
int lastButtonStates[]={HIGH, HIGH, HIGH, HIGH};
int currentButtonStates[]={HIGH, HIGH, HIGH, HIGH};
long lastDebounceTime[4];

int gridCounter = 0;
unsigned long delayCounter = 0;
int b = 1;

boolean blinkOn = true;

// Convert normal decimal numbers to binary coded decimal
byte decToBcd(byte val){
  return ( (val/10*16) + (val%10) );
}

// Convert binary coded decimal to normal decimal numbers
byte bcdToDec(byte val){
  return ( (val/16*10) + (val%16) );
}

void getDateDS3232(){
  // Reset the register pointer
  Wire.beginTransmission(DS3232_I2C_ADDRESS);
  Wire.write(0x00); //move to reg 0
  Wire.endTransmission();

  int vBytesToRead = 7;  
  Wire.requestFrom(DS3232_I2C_ADDRESS, vBytesToRead);
  
  // A few of these need masks because certain bits are control bits
  pD.second     = bcdToDec(Wire.read() & 0x7f);
  pD.minute     = bcdToDec(Wire.read() & 0x7f);
  pD.hour       = bcdToDec(Wire.read() & 0x3f);  // Need to change this if 12 hour am/pm
  pD.dayOfWeek  = bcdToDec(Wire.read() & 0x07); //0= sunday
  pD.dayOfMonth = bcdToDec(Wire.read() & 0x3f);
  pD.month      = bcdToDec(Wire.read() & 0x1f);
  pD.year       = bcdToDec(Wire.read());

  Wire.endTransmission();
}

void checkButtonStates() {
    int idx = 0;
    
    for (int i = 0; i<4; i++) {
        buttonReadings[i] = digitalRead(buttons[i]);
        
        if (buttonReadings[i] != lastButtonStates[i]) lastDebounceTime[i] = millis();
        
        if (millis() - lastDebounceTime[i] > 50) {
            if (currentButtonStates[i] != buttonReadings[i]) {
                currentButtonStates[i] = buttonReadings[i];
                
                //actions
                if (currentButtonStates[i] == LOW) {
                    
                    if (i == IDX_BTN_UP) {
                        if (currentDisplayMode != DATE_DOW) {
                            if (currentSetupMode == BLINK_HOUR) {
                                pD.hour++;
                                if (pD.hour > 23) pD.hour = 0;                            
                            } else if (currentSetupMode == BLINK_MINUTE) {
                                pD.minute++;
                                if (pD.minute > 59) pD.minute = 0;
                            } else if (currentSetupMode == BLINK_SECOND) {
                                pD.second++;
                                if (pD.second > 59) pD.second = 0;
                            } 
                        } else {
                            if (currentSetupMode == BLINK_YEAR) {
                                pD.year++;
                            } else if (currentSetupMode == BLINK_MONTH) {
                                pD.month++;
                                if (pD.month > 12) pD.month = 1;
                            } else if (currentSetupMode == BLINK_DAY) {
                                pD.dayOfMonth++;
                                if (pD.dayOfMonth > 31) pD.dayOfMonth = 1;
                            } else if (currentSetupMode == BLINK_DOW) {
                                pD.dayOfWeek++;
                                if (pD.dayOfWeek > 6) pD.dayOfWeek = 0;
                            }
                        }
                        
                        setDate3232();
                    } else if (i == IDX_BTN_DOWN) {
                        if (currentDisplayMode !=DATE_DOW) {
                            if (currentSetupMode == BLINK_HOUR) {
                                pD.hour--;
                                if (pD.hour < 0) pD.hour = 23;                            
                            } else if (currentSetupMode == BLINK_MINUTE) {
                                pD.minute--;
                                if (pD.minute < 0) pD.minute = 59;
                            } else if (currentSetupMode == BLINK_SECOND) {
                                pD.second--;
                                if (pD.second < 0) pD.second = 59;
                            }
                        } else {
                            if (currentSetupMode == BLINK_YEAR) {
                                pD.year--;
                                if (pD.year < 0) pD.year = 0;                            
                            } else if (currentSetupMode == BLINK_MONTH) {
                                pD.month--;
                                if (pD.month < 1) pD.month = 12;
                            } else if (currentSetupMode == BLINK_DAY) {
                                pD.dayOfMonth--;
                                if (pD.dayOfMonth < 1) pD.dayOfMonth = 31;
                            } else if (currentSetupMode == BLINK_DOW) {
                                pD.dayOfWeek--;
                                if (pD.dayOfWeek < 0) pD.dayOfWeek = 6;
                            }
                        }
                        
                        setDate3232();                        
                    } else if (i == IDX_BTN_DISP_MODE) {
                        currentDisplayMode = (DisplayMode) ((currentDisplayMode + 1) % 3);       
                        currentSetupMode = 0;
                        
                    } else if (i == IDX_BTN_SETUP) {
                        if (currentDisplayMode != DATE_DOW) {//time display
                            currentSetupMode = (currentSetupMode + 1) % 4;
                        } else {
                            currentSetupMode = (currentSetupMode + 1) % 5;                            
                        }                            
                    }
                }
            }
        }
        
        lastButtonStates[i] = buttonReadings[i];
    }               
}

void debugPrint(){
    Serial.print(pD.hour);Serial.print(":");
    Serial.print(pD.minute);Serial.print(":");
    Serial.print(pD.second);Serial.print(" ");
    Serial.print(pD.month);Serial.print("/");
    Serial.print(pD.dayOfMonth);Serial.print("/");
    Serial.println(pD.year);
}

void setDate3232(){
  Wire.beginTransmission(DS3232_I2C_ADDRESS);
  Wire.write(0x00);
  Wire.write(decToBcd(pD.second)); // sec
  Wire.write(decToBcd(pD.minute)); //min
  Wire.write(decToBcd(pD.hour)); //h
  Wire.write(decToBcd(pD.dayOfWeek));  //dow 0= sunday
  Wire.write(decToBcd(pD.dayOfMonth)); //day
  Wire.write(decToBcd(pD.month));  //month
  Wire.write(decToBcd(pD.year)); //y
  Wire.endTransmission(); 
}

void setControlRegisters(){
  Wire.beginTransmission(DS3232_I2C_ADDRESS);
  Wire.write(0x0E);  //Goto register 0Eh
  Wire.write(B00011100);
  Wire.write(B10000000);
  Wire.endTransmission();
}

void setupRTC3232(){
  Wire.begin();
  setControlRegisters();
}

void displayDateTime()
{    
    if (currentDisplayMode != DATE_DOW) {
        switch (gridCounter)
        {
            case HOUR1:
            {
                int hh = pD.hour;
                
                if (currentDisplayMode == TIME_12) {
                    if (hh > 12) hh = hh - 12;                    
                } 
                    
                if (hh < 10) {
                    b = 0;
                } else {
                    b = DIGITS[hh / 10];                
                }
                
                if (currentSetupMode == BLINK_HOUR && !blinkOn) b = 0;                
                break;            
            }
            case HOUR2:
            {
                int hh = pD.hour;
                
                if (currentDisplayMode == TIME_12) {
                    if (hh > 12) hh = hh - 12;  
                    if (hh == 0) hh = 12;                  
                } 
                
                b = DIGITS[(hh - (hh / 10) * 10)];            
                
                if (currentSetupMode == BLINK_HOUR && !blinkOn) b = 0;
                break;            
            }
            case MINUTE1:
            {
                b = DIGITS[pD.minute / 10];
                if (currentSetupMode == BLINK_MINUTE && !blinkOn) b = 0;
                break;        
            }
            case MINUTE2:
            {
                b = DIGITS[(pD.minute - (pD.minute / 10) * 10)];
                if (currentSetupMode == BLINK_MINUTE && !blinkOn) b = 0;
                break;        
            }
            case SECOND1:
            {
                b = DIGITS[pD.second / 10];
                if (currentSetupMode == BLINK_SECOND && !blinkOn) b = 0;                
                break;        
            }
            case SECOND2:
            {
                b = DIGITS[pD.second - (pD.second / 10) * 10];
                if (currentSetupMode == BLINK_SECOND && !blinkOn) b = 0;                
                break;    
            }
            case DASH1:
            case DASH2:            
                b = DASH;
                break;          
            default:
                b = 0;
                break;            
        }
    } else {
        switch (gridCounter)
        {
            case MONTH1:
            {
                if (pD.month < 10) {
                    b = 0;
                } else {
                    b = DIGITS[pD.month / 10];
                }

                if (currentSetupMode == BLINK_MONTH && !blinkOn) b = 0;                  
                break;
            }
            case MONTH2:
            {
                b = DIGITS[pD.month - (pD.month / 10) * 10];
                if (currentSetupMode == BLINK_MONTH && !blinkOn) b = 0;                  
                break;  
            }
            case DAY1:
            {
                b = DIGITS[pD.dayOfMonth / 10];
                if (currentSetupMode == BLINK_DAY && !blinkOn) b = 0;                                  
                break;        
            }
            case DAY2:
            {
                b = DIGITS[pD.dayOfMonth - (pD.dayOfMonth / 10) * 10];
                if (currentSetupMode == BLINK_DAY && !blinkOn) b = 0;                  
                break;
            }
            case YEAR1:
            {
                b = DIGITS[2]; 
                if (currentSetupMode == BLINK_YEAR && !blinkOn) b = 0;                                  
                break;    
            }
            case YEAR2:
            {
                b = DIGITS[(pD.year - (pD.year / 1000) * 1000) / 100];
                if (currentSetupMode == BLINK_YEAR && !blinkOn) b = 0;                  
                if (currentSetupMode == BLINK_YEAR && !blinkOn) b = 0;                  
                break;
            }
            case YEAR3: 
            {
                int d1 = pD.year / 1000;
                int d2 = (pD.year - d1 * 1000) / 100;
                int d3 = (pD.year - d1 * 1000 - d2 * 100) / 10;
                
                b = DIGITS[d3];
                if (currentSetupMode == BLINK_YEAR && !blinkOn) b = 0;                  
                
                break;
            }
            case YEAR4:
            {
                int d1 = pD.year / 1000;
                int d2 = (pD.year - d1 * 1000) / 100;
                int d3 = (pD.year - d1 * 1000 - d2 * 100) / 10;
                int d4 = pD.year - d1 * 1000 - d2 * 100 - d3 * 10;
                
                b = DIGITS[d4];
                if (currentSetupMode == BLINK_YEAR && !blinkOn) b = 0;                  
                
                break;
            }
            case DOW:
            {
                b = DIGITS[pD.dayOfWeek];
                if (currentSetupMode == BLINK_DOW && !blinkOn) b = 0;                  
                break;
            }
            case DASH3:
            case DASH4:
                b = DASH;
                break;
            default:
                b = 0;
                break;                                   
        }
    }
 
    //displays DIGITS[gridCounter]
    digitalWrite(PIN_LATCH, LOW);
    shiftOut(PIN_DS, PIN_CLK, MSBFIRST, b >> 8);
    shiftOut(PIN_DS, PIN_CLK, MSBFIRST, b);
    digitalWrite(PIN_LATCH, HIGH);           
}

void setup()
{
    for (int i = 2; i<=9; i++) pinMode(i, OUTPUT);
    
    pinMode(BTN_DISP_MODE, INPUT);
    digitalWrite(BTN_DISP_MODE, HIGH);

    pinMode(BTN_SETUP, INPUT);
    digitalWrite(BTN_SETUP, HIGH);
    
    pinMode(BTN_UP, INPUT);
    digitalWrite(BTN_UP, HIGH);
    
    pinMode(BTN_DOWN, INPUT);
    digitalWrite(BTN_DOWN, HIGH);
    
    setupRTC3232();
    
    if (digitalRead(BTN_DISP_MODE) == LOW) {
        setDate3232();
    }
    
    Serial.begin(9600);
    Serial.println("DEBUG");        
}

void loop()
{
    //generate pulse to drive each grid 
    digitalWrite(PIN_B0, gridCounter & 1);
    digitalWrite(PIN_B1, (gridCounter & 2) >> 1);
    digitalWrite(PIN_B2, (gridCounter & 4) >> 2);
    digitalWrite(PIN_B3, (gridCounter & 8) >> 3);
    digitalWrite(PIN_OE, LOW);
    
    displayDateTime();
    checkButtonStates(); 
    getDateDS3232(); 
     
    delayCounter++;
    
    if (delayCounter % BLINK_DELAY ==0) blinkOn = !blinkOn;
    //debugPrint();    

    //clears current digit to prevent ghosting
    digitalWrite(PIN_LATCH, LOW);
    shiftOut(PIN_DS, PIN_CLK, MSBFIRST, 0);
    shiftOut(PIN_DS, PIN_CLK, MSBFIRST, 0);
    digitalWrite(PIN_LATCH, HIGH);
  
    delayMicroseconds(30);
    
    digitalWrite(PIN_OE, HIGH);
    digitalWrite(PIN_B0, LOW);
    digitalWrite(PIN_B1, LOW);
    digitalWrite(PIN_B2, LOW);
    digitalWrite(PIN_B3, LOW);  

    gridCounter++;        
    
    //skip the last segment.
    if (gridCounter >= NUM_OF_GRIDS - 1) gridCounter = 0;      
}

